# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test_sort_cuda_basic():
    data = ak.Array([[7, 5, 7], [], [2], [8, 2]])
    gpu_data = ak.to_backend(data, "cuda")
    gpu_sorted = ak.sort(gpu_data)
    result = ak.to_backend(gpu_sorted, "cpu")

    assert to_list(result) == [[5, 7, 7], [], [2], [2, 8]]


def test_sort_cuda_descending():
    data = ak.Array([[3, 1, 2], [5, 4], [], [6]])
    gpu_data = ak.to_backend(data, "cuda")
    gpu_sorted = ak.sort(gpu_data, ascending=False)
    result = ak.to_backend(gpu_sorted, "cpu")

    assert to_list(result) == [[3, 2, 1], [5, 4], [], [6]]


def test_sort_cuda_float():
    data = ak.Array([[3.5, 1.2, 2.8], [5.1, 4.9], [], [6.0]])
    gpu_data = ak.to_backend(data, "cuda")
    gpu_sorted = ak.sort(gpu_data)
    result = ak.to_backend(gpu_sorted, "cpu")

    expected = [[1.2, 2.8, 3.5], [4.9, 5.1], [], [6.0]]
    result_list = to_list(result)

    # Compare with tolerance for floats
    assert len(result_list) == len(expected)
    for res_sublist, exp_sublist in zip(result_list, expected, strict=True):
        assert len(res_sublist) == len(exp_sublist)
        for res_val, exp_val in zip(res_sublist, exp_sublist, strict=True):
            assert abs(res_val - exp_val) < 1e-10


def test_sort_cuda_large():
    # Create random data
    np.random.seed(42)
    data_list = []
    for _ in range(100):
        size = np.random.randint(0, 50)
        if size > 0:
            data_list.append(np.random.randint(0, 100, size).tolist())
        else:
            data_list.append([])

    data = ak.Array(data_list)
    cpu_sorted = ak.sort(data)

    gpu_data = ak.to_backend(data, "cuda")
    gpu_sorted = ak.sort(gpu_data)
    result = ak.to_backend(gpu_sorted, "cpu")

    assert to_list(result) == to_list(cpu_sorted)


def test_sort_cuda_nested():
    data = ak.Array([[[3, 1, 2], [5, 4]], [[9, 7, 8]], [[6]]])
    cpu_sorted = ak.sort(data, axis=-1)

    gpu_data = ak.to_backend(data, "cuda")
    gpu_sorted = ak.sort(gpu_data, axis=-1)
    result = ak.to_backend(gpu_sorted, "cpu")

    assert to_list(result) == to_list(cpu_sorted)
    assert to_list(result) == [[[1, 2, 3], [4, 5]], [[7, 8, 9]], [[6]]]


def test_sort_cuda_deeply_nested():
    data = ak.Array([[[[5, 2, 8], [1, 3]], [[4, 6]]], [[[9, 7]]]])
    cpu_sorted = ak.sort(data, axis=-1)

    gpu_data = ak.to_backend(data, "cuda")
    gpu_sorted = ak.sort(gpu_data, axis=-1)
    result = ak.to_backend(gpu_sorted, "cpu")

    assert to_list(result) == to_list(cpu_sorted)
    assert to_list(result) == [[[[2, 5, 8], [1, 3]], [[4, 6]]], [[[7, 9]]]]


def test_sort_cuda_unsupported_axis():
    """Test that sorting at unsupported axes fails with clear error."""
    # Sorting at axis=-2 requires CuPy kernels that don't exist
    # This should fail with an AssertionError indicating missing kernels
    data = ak.Array([[[7, 2, 3], [4, 5, 6]]])
    gpu_data = ak.to_backend(data, "cuda")

    # axis=-1 should work (our cuda.compute implementation)
    sorted_axis_minus1 = ak.sort(gpu_data, axis=-1)
    result = ak.to_backend(sorted_axis_minus1, "cpu")
    assert to_list(result) == [[[2, 3, 7], [4, 5, 6]]]

    # axis=-2 should fail (requires CuPy kernels not available)
    with pytest.raises(
        AssertionError,
        match=r"(CuPyKernel not found|Operation .* is not supported)",
    ):
        ak.sort(gpu_data, axis=-2)


def test_sort_cuda_no_compute():
    """Test that helpful error is raised when cuda.compute is not available."""
    from awkward._connect.cuda import _compute as cuda_compute

    original_available = cuda_compute._cuda_compute_available

    try:
        # Temporarily make cuda.compute unavailable
        cuda_compute._cuda_compute_available = False

        data = ak.Array([[7, 5, 7], [], [2], [8, 2]])
        gpu_data = ak.to_backend(data, "cuda")

        with pytest.raises(NotImplementedError, match=r"cuda\.compute"):
            ak.sort(gpu_data)

    finally:
        # Restore original state
        cuda_compute._cuda_compute_available = original_available



from __future__ import annotations

import os

import pytest

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")


def simple_test(tmp_path):
    array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    array1 = ak.Array([[1.1, 2.2, 3.3, 4.4], [4.0], [4.4, 5.5]])
    array2 = ak.Array([[1.0, 3.0, 3.3, 4.4], [4.0], [4.4, 10.0], [11.11]])
    ak.to_parquet(
        array, os.path.join(tmp_path, "arr1.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array1, os.path.join(tmp_path, "arr2.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array2, os.path.join(tmp_path, "arr3.parquet"), parquet_compliant_nested=True
    )

    ak.to_parquet_dataset(tmp_path, filenames="arr[1-3].parquet")
    assert os.path.exists(os.path.join(tmp_path, "_common_metadata"))
    assert os.path.exists(os.path.join(tmp_path, "_metadata"))

    with_metadata = ak.from_parquet(tmp_path)
    print(with_metadata.to_list())
    assert with_metadata.tolist() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [1.1, 2.2, 3.3, 4.4],
        [4.0],
        [4.4, 5.5],
        [1.0, 3.0, 3.3, 4.4],
        [4.0],
        [4.4, 10.0],
        [11.11],
    ]


def complex_test(tmp_path):
    array1 = ak.Array(
        [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}, {"x": 3.3, "y": [1, 2, 3]}]
    )
    array2 = ak.Array([{"x": 1.8, "y": [3, 5, 6]}])
    array3 = ak.Array([{"x": 4.4, "y": [1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}])
    ak.to_parquet(
        array1, os.path.join(tmp_path, "arr1.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array2, os.path.join(tmp_path, "arr2.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array3, os.path.join(tmp_path, "arr3.parquet"), parquet_compliant_nested=True
    )

    ak.to_parquet_dataset(tmp_path)

    assert os.path.exists(os.path.join(tmp_path, "_common_metadata"))
    assert os.path.exists(os.path.join(tmp_path, "_metadata"))

    with_metadata = ak.from_parquet(tmp_path)
    assert with_metadata.tolist() == [
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [1, 2]},
        {"x": 3.3, "y": [1, 2, 3]},
        {"x": 1.8, "y": [3, 5, 6]},
        {"x": 4.4, "y": [1, 2, 3, 4]},
        {"x": 5.5, "y": [1, 2, 3, 4, 5]},
    ]


def test_filenames(tmp_path):
    array = ak.Array(
        [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}, {"x": 3.3, "y": [1, 2, 3]}]
    )
    array1 = ak.Array([{"x": 1.8, "y": [3, 5, 6]}])
    array2 = ak.Array([{"x": 4.4, "y": [1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}])
    ak.to_parquet(
        array, os.path.join(tmp_path, "arr1.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array1, os.path.join(tmp_path, "arr2.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array2, os.path.join(tmp_path, "arr3.parquet"), parquet_compliant_nested=True
    )

    ak.to_parquet_dataset(
        tmp_path, filenames=["arr1.parquet", "arr2.parquet", "arr3.parquet"]
    )

    assert os.path.exists(os.path.join(tmp_path, "_common_metadata"))
    assert os.path.exists(os.path.join(tmp_path, "_metadata"))

    with_metadata = ak.from_parquet(tmp_path)
    assert with_metadata.tolist() == [
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [1, 2]},
        {"x": 3.3, "y": [1, 2, 3]},
        {"x": 1.8, "y": [3, 5, 6]},
        {"x": 4.4, "y": [1, 2, 3, 4]},
        {"x": 5.5, "y": [1, 2, 3, 4, 5]},
    ]


def test_wildcard(tmp_path):
    array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    array1 = ak.Array([[1.1, 2.2, 3.3, 4.4], [4.0], [4.4, 5.5]])
    array2 = ak.Array([[1.0, 3.0, 3.3, 4.4], [4.0], [4.4, 10.0], [11.11]])
    ak.to_parquet(
        array, os.path.join(tmp_path, "arr1.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array1, os.path.join(tmp_path, "arr2.parquet"), parquet_compliant_nested=True
    )
    ak.to_parquet(
        array2, os.path.join(tmp_path, "arr3.parquet"), parquet_compliant_nested=True
    )

    ak.to_parquet_dataset(tmp_path, filenames="arr?.parquet")

    with_metadata = ak.from_parquet(tmp_path)
    assert with_metadata.tolist() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [1.1, 2.2, 3.3, 4.4],
        [4.0],
        [4.4, 5.5],
        [1.0, 3.0, 3.3, 4.4],
        [4.0],
        [4.4, 10.0],
        [11.11],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak

numba = pytest.importorskip("numba")
jax = pytest.importorskip("jax")


def test():
    behavior = {}

    ak.jax.register_and_check()

    input_arr = ak.Array([1.0], backend="jax")

    @numba.vectorize(
        [
            numba.float32(numba.float32, numba.float32),
            numba.float64(numba.float64, numba.float64),
        ]
    )
    def _some_kernel(x, y):
        return x * x + y * y

    @ak.mixin_class(behavior)
    class SomeClass:
        @property
        def some_kernel(self):
            return _some_kernel(self.x, self.y)

    ak.behavior.update(behavior)

    arr = ak.zip({"x": input_arr, "y": input_arr}, with_name="SomeClass")

    assert ak.all(arr.some_kernel == ak.Array([2.0], backend="jax"))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from types import ModuleType

from packaging.version import parse as parse_version

__all__ = [
    "AwkwardArrowArray",
    "AwkwardArrowType",
    "and_validbytes",
    "convert_awkward_arrow_table_to_native",
    "convert_native_arrow_table_to_awkward",
    "convert_to_array",
    "direct_Content_subclass",
    "direct_Content_subclass_name",
    "form_handle_arrow",
    "handle_arrow",
    "import_pyarrow",
    "import_pyarrow_compute",
    "import_pyarrow_parquet",
    "popbuffers",
    "remove_optiontype",
    "to_awkwardarrow_storage_types",
    "to_awkwardarrow_type",
    "to_length",
    "to_null_count",
    "to_validbits",
]

try:
    import pyarrow

    error_message = None

except ModuleNotFoundError:
    pyarrow = None
    error_message = """to use {0}, you must install pyarrow:

    pip install pyarrow

or

    conda install -c conda-forge pyarrow
"""

else:
    if parse_version(pyarrow.__version__) < parse_version("7.0.0"):
        pyarrow = None
        error_message = "pyarrow 7.0.0 or later required for {0}"

if error_message is None:
    from .conversions import (
        and_validbytes,
        convert_to_array,
        direct_Content_subclass,
        direct_Content_subclass_name,
        form_handle_arrow,
        handle_arrow,
        popbuffers,
        remove_optiontype,
        to_awkwardarrow_type,
        to_length,
        to_null_count,
        to_validbits,
    )
    from .extn_types import (
        AwkwardArrowArray,
        AwkwardArrowType,
        to_awkwardarrow_storage_types,
    )
    from .table_conv import (
        convert_awkward_arrow_table_to_native,
        convert_native_arrow_table_to_awkward,
    )
else:
    AwkwardArrowArray = None
    AwkwardArrowType = None

    def nothing_without_pyarrow(*args, **kwargs):
        raise NotImplementedError(
            "This function requires pyarrow, which is not installed."
        )

    convert_awkward_arrow_table_to_native = nothing_without_pyarrow
    convert_native_arrow_table_to_awkward = nothing_without_pyarrow
    and_validbytes = nothing_without_pyarrow
    to_validbits = nothing_without_pyarrow
    to_length = nothing_without_pyarrow
    to_null_count = nothing_without_pyarrow
    to_awkwardarrow_storage_types = nothing_without_pyarrow
    popbuffers = nothing_without_pyarrow
    handle_arrow = nothing_without_pyarrow
    convert_to_array = nothing_without_pyarrow
    to_awkwardarrow_type = nothing_without_pyarrow
    direct_Content_subclass = nothing_without_pyarrow
    direct_Content_subclass_name = nothing_without_pyarrow
    remove_optiontype = nothing_without_pyarrow
    form_handle_arrow = nothing_without_pyarrow


def import_pyarrow(name: str) -> ModuleType:
    if pyarrow is None:
        raise ImportError(error_message.format(name))
    return pyarrow


def import_pyarrow_parquet(name: str) -> ModuleType:
    if pyarrow is None:
        raise ImportError(error_message.format(name))

    import pyarrow.parquet as out

    return out


def import_pyarrow_compute(name: str) -> ModuleType:
    if pyarrow is None:
        raise ImportError(error_message.format(name))

    import pyarrow.compute as out

    return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import re

import pytest

import awkward as ak


@pytest.mark.parametrize("forget_length", [False, True])
def test_invalid(forget_length):
    layout = ak.contents.RecordArray(
        [
            ak.contents.NumpyArray([1, 2, 3]),
            ak.contents.NumpyArray([1, 2, 3]),
        ],
        ["x", "x"],
    )
    assert re.match(r".*duplicate field 'x'.*", ak.validity_error(layout)) is not None
    assert (
        re.match(
            r".*duplicate field 'x'.*",
            ak.validity_error(layout.to_typetracer(forget_length)),
        )
        is not None
    )


@pytest.mark.parametrize("forget_length", [False, True])
def test_valid(forget_length):
    layout = ak.contents.RecordArray(
        [
            ak.contents.NumpyArray([1, 2, 3]),
            ak.contents.NumpyArray([1, 2, 3]),
        ],
        ["x", "y"],
    )
    assert re.match(r".*duplicate field 'x'.*", ak.validity_error(layout)) is None
    assert (
        re.match(
            r".*duplicate field 'x'.*",
            ak.validity_error(layout.to_typetracer(forget_length)),
        )
        is None
    )



from __future__ import annotations

import numpy as np

import awkward as ak
from awkward.typetracer import unknown_length


def test():
    arr = (
        ak.mask([1, 2, 3], [True, False, False], highlevel=False)
        .to_BitMaskedArray(valid_when=True, lsb_order=True)
        .to_typetracer(forget_length=True)
    )

    result = arr.mask_as_bool()
    assert result.dtype == np.dtype("bool")
    assert result.ndim == 1
    assert result.size is unknown_length



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

to_raggedtensor = ak.operations.to_raggedtensor
from_raggedtensor = ak.operations.from_raggedtensor

tf = pytest.importorskip("tensorflow")

content = ak.contents.NumpyArray(
    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
)
starts1 = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
stops1 = ak.index.Index64(np.array([3, 3, 5, 6, 9]))
starts2 = ak.index.Index64(np.array([0, 3]))
stops2 = ak.index.Index64(np.array([3, 5]))

array = np.arange(2 * 3 * 5).reshape(2, 3, 5)
content2 = ak.contents.NumpyArray(array.reshape(-1))
inneroffsets = ak.index.Index64(np.array([0, 5, 10, 15, 20, 25, 30]))
outeroffsets = ak.index.Index64(np.array([0, 3, 6]))


def test_convert_to_raggedtensor():
    # a test for ListArray -> RaggedTensor
    array1 = ak.contents.ListArray(starts1, stops1, content)
    assert to_raggedtensor(array1).to_list() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    # a test for awkward.highlevel.Array -> RaggedTensor
    array2 = ak.Array(array1)
    assert to_raggedtensor(array2).to_list() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]

    # a test for NumpyArray -> RaggedTensor
    array3 = content
    assert to_raggedtensor(array3).to_list() == [
        [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]
    ]

    # a test for RegularArray -> RaggedTensor
    array4 = ak.contents.RegularArray(content, size=2)
    assert to_raggedtensor(array4).to_list() == [
        [1.1, 2.2],
        [3.3, 4.4],
        [5.5, 6.6],
        [7.7, 8.8],
    ]

    # try a single line awkward array
    array5 = ak.Array([3, 1, 4, 1, 9, 2, 6])
    assert to_raggedtensor(array5).to_list() == [[3, 1, 4, 1, 9, 2, 6]]

    # try a multiple ragged array
    array6 = ak.Array([[[1.1, 2.2], [3.3]], [], [[4.4, 5.5]]])
    assert to_raggedtensor(array6).to_list() == [[[1.1, 2.2], [3.3]], [], [[4.4, 5.5]]]

    # try a listoffset array inside a listoffset array
    array7 = ak.contents.ListOffsetArray(
        outeroffsets, ak.contents.ListOffsetArray(inneroffsets, content2)
    )
    assert to_raggedtensor(array7).to_list() == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    # try a list array inside a list array

    array8 = ak.contents.ListArray(
        starts2, stops2, ak.contents.ListArray(starts1, stops1, content)
    )
    assert to_raggedtensor(array8).to_list() == [
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]],
        [[6.6], [7.7, 8.8, 9.9]],
    ]

    # try just a python list
    array9 = [3, 1, 4, 1, 9, 2, 6]
    assert to_raggedtensor(array9).to_list() == [[3, 1, 4, 1, 9, 2, 6]]


np_array1 = np.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float32)

offsets1 = ak.index.Index64(np.array([0, 2, 3, 3, 5]))
content1 = ak.contents.NumpyArray(np_array1)


def test_convert_from_raggedtensor():
    tf_array1 = tf.RaggedTensor.from_row_splits(
        values=[1.1, 2.2, 3.3, 4.4, 5.5], row_splits=[0, 2, 3, 3, 5]
    )

    ak_array1 = ak.contents.ListOffsetArray(offsets1, content1)
    result1 = ak.to_layout(from_raggedtensor(tf_array1), allow_record=False)
    assert (
        result1.content.data == ak.to_backend(np_array1, result1.backend).layout.data
    ).all()
    assert (
        result1.offsets.data
        == ak.to_backend([0, 2, 3, 3, 5], result1.backend).layout.data
    ).all()
    assert from_raggedtensor(tf_array1).to_list() == ak_array1.to_list()

    tf_array2 = tf.RaggedTensor.from_nested_row_splits(
        flat_values=[3, 1, 4, 1, 5, 9, 2, 6],
        nested_row_splits=([0, 3, 3, 5], [0, 4, 4, 7, 8, 8]),
    )
    assert from_raggedtensor(tf_array2).to_list() == [
        [[3, 1, 4, 1], [], [5, 9, 2]],
        [],
        [[6], []],
    ]



# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

@pytest.mark.skip(reason='Unable to generate any tests for kernel')
def test_cudaawkward_RegularArray_combinations_64_1():
    raise NotImplementedError('Unable to generate any tests for kernel')



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak


def test_from_numpy_attrs():
    """Test that ak.from_numpy correctly passes attrs parameter."""
    array = np.array([1, 2, 3])
    result = ak.from_numpy(array, attrs={"hello": "world"})
    assert result.attrs == {"hello": "world"}
    assert result.to_list() == [1, 2, 3]


def test_from_numpy_attrs_none():
    """Test that ak.from_numpy works with attrs=None."""
    array = np.array([1, 2, 3])
    result = ak.from_numpy(array, attrs=None)
    assert result.attrs == {}
    assert result.to_list() == [1, 2, 3]


def test_from_numpy_attrs_with_structured_array():
    """Test that ak.from_numpy passes attrs with structured arrays."""
    array = np.array([(1, 2.5), (3, 4.5)], dtype=[("x", "i4"), ("y", "f8")])
    result = ak.from_numpy(array, attrs={"source": "numpy"})
    assert result.attrs == {"source": "numpy"}
    assert ak.fields(result) == ["x", "y"]


def test_from_numpy_attrs_with_behavior():
    """Test that ak.from_numpy works with both attrs and behavior."""
    array = np.array([1, 2, 3])
    result = ak.from_numpy(
        array, attrs={"hello": "world"}, behavior={"custom": "behavior"}
    )
    assert result.attrs == {"hello": "world"}
    assert result.behavior == {"custom": "behavior"}


def test_from_cupy_attrs():
    """Test that ak.from_cupy correctly passes attrs parameter."""
    cupy = pytest.importorskip("cupy")
    array = cupy.array([1, 2, 3])
    result = ak.from_cupy(array, attrs={"foo": "bar"})
    assert result.attrs == {"foo": "bar"}
    assert result.to_list() == [1, 2, 3]


def test_from_cupy_attrs_none():
    """Test that ak.from_cupy works with attrs=None."""
    cupy = pytest.importorskip("cupy")
    array = cupy.array([1, 2, 3])
    result = ak.from_cupy(array, attrs=None)
    assert result.attrs == {}
    assert result.to_list() == [1, 2, 3]


def test_from_jax_attrs():
    """Test that ak.from_jax correctly passes attrs parameter."""
    jax = pytest.importorskip("jax")
    array = jax.numpy.array([1, 2, 3])
    result = ak.from_jax(array, attrs={"jax": "test"})
    assert result.attrs == {"jax": "test"}
    assert result.to_list() == [1, 2, 3]


def test_from_jax_attrs_none():
    """Test that ak.from_jax works with attrs=None."""
    jax = pytest.importorskip("jax")
    array = jax.numpy.array([1, 2, 3])
    result = ak.from_jax(array, attrs=None)
    assert result.attrs == {}
    assert result.to_list() == [1, 2, 3]


def test_from_dlpack_attrs():
    """Test that ak.from_dlpack correctly passes attrs parameter."""
    # DLPack is supported by NumPy 1.22+
    array = np.array([1, 2, 3])
    if not hasattr(array, "__dlpack__"):
        pytest.skip("NumPy version doesn't support DLPack")

    result = ak.from_dlpack(array, attrs={"dlpack": "test"})
    assert result.attrs == {"dlpack": "test"}
    assert result.to_list() == [1, 2, 3]


def test_from_dlpack_attrs_none():
    """Test that ak.from_dlpack works with attrs=None."""
    array = np.array([1, 2, 3])
    if not hasattr(array, "__dlpack__"):
        pytest.skip("NumPy version doesn't support DLPack")

    result = ak.from_dlpack(array, attrs=None)
    assert result.attrs == {}
    assert result.to_list() == [1, 2, 3]


def test_from_numpy_attrs_multidimensional():
    """Test that attrs work with multidimensional arrays."""
    array = np.array([[1, 2], [3, 4]])
    result = ak.from_numpy(array, regulararray=True, attrs={"shape": "2x2"})
    assert result.attrs == {"shape": "2x2"}
    assert result.to_list() == [[1, 2], [3, 4]]


def test_from_numpy_attrs_complex():
    """Test that attrs work with complex nested attributes."""
    array = np.array([1, 2, 3])
    complex_attrs = {"meta": {"version": "1.0", "source": "test"}, "id": 42}
    result = ak.from_numpy(array, attrs=complex_attrs)
    assert result.attrs == complex_attrs
    assert result.to_list() == [1, 2, 3]


def test_from_functions_attrs_preserved_in_operations():
    """Test that attrs from from_* functions are preserved in operations."""
    array = np.array([1, 2, 3])
    result = ak.from_numpy(array, attrs={"original": "data"})

    # Test that attrs survive slicing
    sliced = result[1:]
    assert sliced.attrs == {"original": "data"}
    assert sliced.to_list() == [2, 3]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak


def test_can_cast():
    numpy_like = ak._nplikes.numpy.Numpy.instance()
    # default casting is "same_kind"
    assert numpy_like.can_cast(np.float32, np.float64)
    assert numpy_like.can_cast(np.int32, np.int64)
    assert numpy_like.can_cast(np.float64, np.float32)
    assert numpy_like.can_cast(np.int64, np.int32)

    assert numpy_like.can_cast(np.float32, np.float64, casting="safe")
    assert numpy_like.can_cast(np.int32, np.int64, casting="safe")
    assert not numpy_like.can_cast(np.float64, np.float32, casting="safe")
    assert not numpy_like.can_cast(np.int64, np.int32, casting="safe")

    assert numpy_like.can_cast(np.float32, np.float32, casting="equiv")
    assert not numpy_like.can_cast(np.float32, np.float64, casting="equiv")
    assert not numpy_like.can_cast(np.int32, np.int64, casting="equiv")
    assert not numpy_like.can_cast(np.float64, np.float32, casting="equiv")
    assert not numpy_like.can_cast(np.int64, np.int32, casting="equiv")


def test_mergeable_next_mergecastable():
    array1 = ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int32))
    array2 = ak.contents.NumpyArray(np.array([4, 5, 6], dtype=np.int64))
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="same_kind")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="family")
    assert not array1._mergeable_next(array2, mergebool=False, mergecastable="equiv")

    array1 = ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype=np.float32))
    array2 = ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64))
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="same_kind")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="family")
    assert not array1._mergeable_next(array2, mergebool=False, mergecastable="equiv")

    array1 = ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int32))
    array2 = ak.contents.NumpyArray(np.array([4, 5, 6], dtype=np.int32))
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="same_kind")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="family")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="equiv")

    array1 = ak.contents.NumpyArray(np.array([1.0, 2.0, 3.0], dtype=np.float32))
    array2 = ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float32))
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="same_kind")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="family")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="equiv")

    array1 = ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64))
    array2 = ak.contents.NumpyArray(np.array([4.0, 5.0, 6.0], dtype=np.float64))
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="same_kind")
    assert not array1._mergeable_next(array2, mergebool=False, mergecastable="family")
    assert not array1._mergeable_next(array2, mergebool=False, mergecastable="equiv")

    array1 = ak.contents.NumpyArray(np.array([True, False, True], dtype=np.bool_))
    array2 = ak.contents.NumpyArray(np.array([False, True, False], dtype=np.bool_))
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="same_kind")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="family")
    assert array1._mergeable_next(array2, mergebool=False, mergecastable="equiv")
    assert array1._mergeable_next(array2, mergebool=True, mergecastable="same_kind")
    assert array1._mergeable_next(array2, mergebool=True, mergecastable="family")
    assert array1._mergeable_next(array2, mergebool=True, mergecastable="equiv")



from __future__ import annotations

from itertools import product

import cupy as cp
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0546_fill_none_replacement_value_type():
    array = ak.operations.values_astype(
        ak.highlevel.Array([1.1, 2.2, None, 3.3]), np.float32
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert str(ak.operations.fill_none(cuda_array, cp.float32(0)).type) == "4 * float32"


def test_0912_packed_indexed_option_array():
    index = ak.index.Index64(np.r_[0, -1, 2, -1, 4])
    content = ak.contents.NumpyArray(np.arange(8))
    layout = ak.contents.IndexedOptionArray(index, content)
    cuda_layout = ak.to_backend(layout, "cuda")

    cuda_packed = ak.operations.to_packed(cuda_layout, highlevel=False)

    assert to_list(cuda_layout) == to_list(cuda_packed)
    assert isinstance(cuda_packed, ak.contents.IndexedOptionArray)
    assert cp.asarray(cuda_packed.index).tolist() == [0, -1, 1, -1, 2]
    assert len(cuda_packed.content) == 3


def test_0912_packed_indexed_array():
    index = ak.index.Index64(np.array([0, 1, 2, 3, 6, 7, 8]))
    content = ak.contents.NumpyArray(np.arange(10))
    layout = ak.contents.IndexedArray(index, content)
    cuda_layout = ak.to_backend(layout, "cuda")

    cuda_packed = ak.operations.to_packed(cuda_layout, highlevel=False)
    assert to_list(cuda_packed) == to_list(cuda_layout)

    assert isinstance(cuda_packed, ak.contents.NumpyArray)
    assert len(cuda_packed) == len(index)


def test_0912_packed_bit_masked_aray():
    mask = ak.index.IndexU8(np.array([0b10101010]))
    content = ak.contents.NumpyArray(np.arange(16))
    layout = ak.contents.BitMaskedArray(mask, content, False, 8, False)
    cuda_layout = ak.to_backend(layout, "cuda")

    cuda_packed = ak.operations.to_packed(cuda_layout, highlevel=False)
    assert to_list(cuda_packed) == to_list(cuda_layout)
    assert len(cuda_packed.content) == 8


def test_0912_packed_union_array():
    a = ak.contents.NumpyArray(np.arange(4))
    b = ak.contents.NumpyArray(np.arange(4) + 4)
    c = ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(12)), 3)
    layout = ak.contents.UnionArray.simplified(
        ak.index.Index8([1, 1, 2, 2, 0, 0]),
        ak.index.Index64([0, 1, 0, 1, 0, 1]),
        [a, b, c],
    )
    cuda_layout = ak.to_backend(layout, "cuda")

    cuda_packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(cuda_packed) == to_list(cuda_layout)
    # Check that it merges like contents
    assert len(cuda_packed.contents) == 2
    index_0 = cp.asarray(cuda_packed.index)[cp.asarray(cuda_packed.tags) == 0]
    assert index_0.tolist() == [0, 1, 2, 3]


empty = ak.highlevel.Array(
    ak.contents.RegularArray(
        ak.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout, 0, zeros_length=0
    )
)
cuda_empty = ak.to_backend(empty, "cuda")


def test_0590_allow_regulararray_size_zero_carry():
    assert cuda_empty[[]].to_list() == []


def test_0590_allow_regulararray_size_zero_mergeable():
    assert ak.operations.concatenate([cuda_empty, cuda_empty]).to_list() == []


def test_0590_allow_regulararray_size_zero_localindex():
    assert ak.operations.local_index(cuda_empty, axis=0).to_list() == []
    assert ak.operations.local_index(cuda_empty, axis=1).to_list() == []
    assert ak.operations.local_index(cuda_empty, axis=2).to_list() == []


def test_0627_behavior_from_dict_of_arrays():
    simple = {"what": "ever"}
    one = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], behavior=simple)
    two = ak.Array([["one", "two"], ["three"], ["four", "five"]], behavior=simple)
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    cuda_three = ak.operations.cartesian({"one": cuda_one, "two": cuda_two})
    assert cuda_three.behavior == {"what": "ever"}


def test_0713_getitem_field_should_simplify_optiontype():
    arr1 = ak.highlevel.Array({"a": [1, 2], "b": [1, None]})
    cuda_arr1 = ak.to_backend(arr1, "cuda")
    cuda_arr2 = ak.operations.mask(cuda_arr1, cp.array([True, True]))

    assert isinstance(cuda_arr2.layout, ak.contents.ByteMaskedArray)
    assert isinstance(cuda_arr2.layout.content, ak.contents.RecordArray)
    assert isinstance(cuda_arr2.layout.content["b"], ak.contents.IndexedOptionArray)

    assert isinstance(cuda_arr2.b.layout, ak.contents.IndexedOptionArray)
    assert isinstance(cuda_arr2.b.layout.content, ak.contents.NumpyArray)

    assert ak.operations.is_none(cuda_arr2.b).to_list() == [False, True]


def test_0766_prevent_combinations_of_characters_cartesian():
    one = ak.Array([1, 2, 3, 4])
    two = ak.Array(["aa", "bb", "cc", "dd"])
    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    with pytest.raises(ValueError):
        ak.operations.cartesian([cuda_one, cuda_two], axis=1)

    two = ak.Array([["aa", "bb"], ["cc"], [], ["dd"]])
    cuda_two = ak.to_backend(two, "cuda")

    assert to_list(ak.operations.cartesian([cuda_one, cuda_two], axis=1)) == [
        [(1, "aa"), (1, "bb")],
        [(2, "cc")],
        [],
        [(4, "dd")],
    ]
    with pytest.raises(ValueError):
        ak.operations.cartesian([cuda_one, cuda_two], axis=2)


def test_0773_typeparser_arraytype_12():
    array = ak.Array([[1, 2], [3, 4], [5, 6]])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(ak.to_regular(cuda_array).type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_0773_typeparser_arraytype_13():
    array = ak.Array([[1, 2], [3, 4], [5, 6]])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(ak.with_parameter(ak.to_regular(cuda_array), "wonky", "string").type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_0773_typeparser_arraytype_14():
    array = ak.Array([1, 2, 3, [1], [1, 2], [1, 2, 3]])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(ak.with_parameter(cuda_array, "wonky", "string").type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_0773_typeparser_arraytype_15():
    array = ak.Array([1, 2, 3, None, [1], [1, 2], [1, 2, 3]])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(ak.with_parameter(cuda_array, "wonky", "string").type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_0773_typeparser_arraytype_16():
    array = ak.Array([1, 2, 3, None, [1], [1, 2], [1, 2, 3]])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(cuda_array.type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_0773_typeparser_arraytype_17():
    array = ak.Array([1, 2, 3, None, [], [], []])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(cuda_array.type)
    parsedtype = ak.types.from_datashape(text, highlevel=False)
    assert str(parsedtype) == text


def test_0773_typeparser_arraytype_categorical_1():
    pytest.importorskip("pyarrow")

    array = ak.Array(["one", "one", "two", "three", "one", "three"])
    cuda_array = ak.to_backend(array, "cuda")

    text = str(ak.str.to_categorical(cuda_array).type)
    parsedtype = ak.types.from_datashape(text, highlevel=True)
    assert isinstance(parsedtype, ak.types.ArrayType)
    assert str(parsedtype) == text


def test_0866_getitem_field_and_flatten_unions_getitem_field():
    a1 = ak.operations.zip(
        {"a": [[1], [], [2, 3]], "b": [[4], [], [5, 6]]}, with_name="a1"
    )
    a2 = ak.operations.zip(
        {"a": [[7, 8], [9], []], "b": [[10, 11], [12], []]}, with_name="a2"
    )
    cuda_a1 = ak.to_backend(a1, "cuda")
    cuda_a2 = ak.to_backend(a2, "cuda")

    union = ak.operations.where(cp.array([True, False, True]), cuda_a1, cuda_a2)
    assert str(union.a.type) == "3 * var * int64"


def test_0898_unzip_heterogeneous_records():
    a = ak.Array([{"this": 100}])
    b = ak.Array([{"this": 90, "that": 100}])
    cuda_a = ak.to_backend(a, "cuda")
    cuda_b = ak.to_backend(b, "cuda")

    cuda_c = ak.operations.concatenate((cuda_a, cuda_b))

    with pytest.raises(ValueError):
        ak.operations.unzip(cuda_c)

    a = ak.Array([{"this": 100}])
    b = ak.Array([{"this": 90}])
    cuda_a = ak.to_backend(a, "cuda")
    cuda_b = ak.to_backend(b, "cuda")
    cuda_c = ak.operations.concatenate((cuda_a, cuda_b))

    (tmp,) = ak.operations.unzip(cuda_c)

    assert tmp.to_list() == [100, 90]


def test_0896_content_classes_refactoring_BitMaskedArray_NumpyArray():
    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)

    assert cuda_a.to_typetracer().form == cuda_a.form
    assert cuda_a.to_typetracer().form.type == cuda_a.form.type
    assert len(cuda_a) == 13
    with pytest.raises(IndexError):
        cuda_a[13]
    with pytest.raises(IndexError):
        cuda_a[-14]
    assert cuda_a[0] == 0.0
    assert cuda_a[1] == 1.0
    assert cuda_a[2] == 2.0
    assert cuda_a[3] == 3.0
    assert cuda_a[4] is None
    assert cuda_a[5] is None
    assert cuda_a[6] is None
    assert cuda_a[7] is None
    assert cuda_a[8] == 1.1
    assert cuda_a[9] is None
    assert cuda_a[10] == 3.3
    assert cuda_a[11] is None
    assert cuda_a[12] == 5.5
    assert cuda_a[-13] == 0.0
    assert cuda_a[-12] == 1.0
    assert cuda_a[-11] == 2.0
    assert cuda_a[-10] == 3.0
    assert cuda_a[-9] is None
    assert cuda_a[-8] is None
    assert cuda_a[-7] is None
    assert cuda_a[-6] is None
    assert cuda_a[-5] == 1.1
    assert cuda_a[-4] is None
    assert cuda_a[-3] == 3.3
    assert cuda_a[-2] is None
    assert cuda_a[-1] == 5.5
    assert isinstance(cuda_a[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_a.to_typetracer()[5:].form == cuda_a[5:].form
    assert len(cuda_a[5:]) == 8
    assert len(cuda_a[-8:]) == 8
    assert len(cuda_a[5:100]) == 8
    assert len(cuda_a[-8:100]) == 8
    assert cuda_a[5:][2] is None
    assert cuda_a[5:][3] == 1.1
    assert cuda_a[-8:][2] is None
    assert cuda_a[-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_a["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    cuda_b = ak.to_backend(b, "cuda", highlevel=False)

    assert cuda_b.to_typetracer().form == cuda_b.form
    assert cuda_b.to_typetracer().form.type == cuda_b.form.type
    assert len(cuda_b) == 13
    with pytest.raises(IndexError):
        cuda_b[13]
    with pytest.raises(IndexError):
        cuda_b[-14]
    assert cuda_b[0] == 0.0
    assert cuda_b[1] == 1.0
    assert cuda_b[2] == 2.0
    assert cuda_b[3] == 3.0
    assert cuda_b[4] is None
    assert cuda_b[5] is None
    assert cuda_b[6] is None
    assert cuda_b[7] is None
    assert cuda_b[8] == 1.1
    assert cuda_b[9] is None
    assert cuda_b[10] == 3.3
    assert cuda_b[11] is None
    assert cuda_b[12] == 5.5
    assert cuda_b[-13] == 0.0
    assert cuda_b[-12] == 1.0
    assert cuda_b[-11] == 2.0
    assert cuda_b[-10] == 3.0
    assert cuda_b[-9] is None
    assert cuda_b[-8] is None
    assert cuda_b[-7] is None
    assert cuda_b[-6] is None
    assert cuda_b[-5] == 1.1
    assert cuda_b[-4] is None
    assert cuda_b[-3] == 3.3
    assert cuda_b[-2] is None
    assert cuda_b[-1] == 5.5
    assert isinstance(cuda_b[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_b.to_typetracer()[5:].form == cuda_b[5:].form
    assert len(cuda_b[5:]) == 8
    assert len(cuda_b[-8:]) == 8
    assert len(cuda_b[5:100]) == 8
    assert len(cuda_b[-8:100]) == 8
    assert cuda_b[5:][2] is None
    assert cuda_b[5:][3] == 1.1
    assert cuda_b[-8:][2] is None
    assert cuda_b[-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_b["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    cuda_c = ak.to_backend(c, "cuda", highlevel=False)

    assert cuda_c.to_typetracer().form == cuda_c.form
    assert cuda_c.to_typetracer().form.type == cuda_c.form.type
    assert len(cuda_c) == 13
    with pytest.raises(IndexError):
        cuda_c[13]
    with pytest.raises(IndexError):
        cuda_c[-14]
    assert cuda_c[0] == 0.0
    assert cuda_c[1] == 1.0
    assert cuda_c[2] == 2.0
    assert cuda_c[3] == 3.0
    assert cuda_c[4] is None
    assert cuda_c[5] is None
    assert cuda_c[6] is None
    assert cuda_c[7] is None
    assert cuda_c[8] == 1.1
    assert cuda_c[9] is None
    assert cuda_c[10] == 3.3
    assert cuda_c[11] is None
    assert cuda_c[12] == 5.5
    assert cuda_c[-13] == 0.0
    assert cuda_c[-12] == 1.0
    assert cuda_c[-11] == 2.0
    assert cuda_c[-10] == 3.0
    assert cuda_c[-9] is None
    assert cuda_c[-8] is None
    assert cuda_c[-7] is None
    assert cuda_c[-6] is None
    assert cuda_c[-5] == 1.1
    assert cuda_c[-4] is None
    assert cuda_c[-3] == 3.3
    assert cuda_c[-2] is None
    assert cuda_c[-1] == 5.5
    assert isinstance(cuda_c[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_c.to_typetracer()[5:].form == cuda_c[5:].form
    assert len(cuda_c[5:]) == 8
    assert len(cuda_c[-8:]) == 8
    assert len(cuda_c[5:100]) == 8
    assert len(cuda_c[-8:100]) == 8
    assert cuda_c[5:][2] is None
    assert cuda_c[5:][3] == 1.1
    assert cuda_c[-8:][2] is None
    assert cuda_c[-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_c["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    cuda_d = ak.to_backend(d, "cuda", highlevel=False)

    assert cuda_d.to_typetracer().form == cuda_d.form
    assert cuda_d.to_typetracer().form.type == cuda_d.form.type
    assert len(cuda_d) == 13
    with pytest.raises(IndexError):
        cuda_d[13]
    with pytest.raises(IndexError):
        cuda_d[-14]
    assert cuda_d[0] == 0.0
    assert cuda_d[1] == 1.0
    assert cuda_d[2] == 2.0
    assert cuda_d[3] == 3.0
    assert cuda_d[4] is None
    assert cuda_d[5] is None
    assert cuda_d[6] is None
    assert cuda_d[7] is None
    assert cuda_d[8] == 1.1
    assert cuda_d[9] is None
    assert cuda_d[10] == 3.3
    assert cuda_d[11] is None
    assert cuda_d[12] == 5.5
    assert cuda_d[-13] == 0.0
    assert cuda_d[-12] == 1.0
    assert cuda_d[-11] == 2.0
    assert cuda_d[-10] == 3.0
    assert cuda_d[-9] is None
    assert cuda_d[-8] is None
    assert cuda_d[-7] is None
    assert cuda_d[-6] is None
    assert cuda_d[-5] == 1.1
    assert cuda_d[-4] is None
    assert cuda_d[-3] == 3.3
    assert cuda_d[-2] is None
    assert cuda_d[-1] == 5.5
    assert isinstance(cuda_d[5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_d.to_typetracer()[5:].form == cuda_d[5:].form
    assert len(cuda_d[5:]) == 8
    assert len(cuda_d[-8:]) == 8
    assert len(cuda_d[5:100]) == 8
    assert len(cuda_d[-8:100]) == 8
    assert cuda_d[5:][2] is None
    assert cuda_d[5:][3] == 1.1
    assert cuda_d[-8:][2] is None
    assert cuda_d[-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_d["bad"]


def test_0896_content_classes_refactoring_BitMaskedArray_RecordArray_NumpyArray():
    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        True,
                        True,
                        True,
                        True,
                        False,
                        False,
                        False,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                    ]
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)

    assert cuda_a.to_typetracer().form == cuda_a.form
    assert cuda_a.to_typetracer().form.type == cuda_a.form.type
    assert len(cuda_a["nest"]) == 13
    assert cuda_a.to_typetracer()["nest"].form == cuda_a["nest"].form
    with pytest.raises(IndexError):
        cuda_a["nest"][13]
    with pytest.raises(IndexError):
        cuda_a["nest"][-14]
    assert cuda_a["nest"][0] == 0.0
    assert cuda_a["nest"][1] == 1.0
    assert cuda_a["nest"][2] == 2.0
    assert cuda_a["nest"][3] == 3.0
    assert cuda_a["nest"][4] is None
    assert cuda_a["nest"][5] is None
    assert cuda_a["nest"][6] is None
    assert cuda_a["nest"][7] is None
    assert cuda_a["nest"][8] == 1.1
    assert cuda_a["nest"][9] is None
    assert cuda_a["nest"][10] == 3.3
    assert cuda_a["nest"][11] is None
    assert cuda_a["nest"][12] == 5.5
    assert cuda_a["nest"][-13] == 0.0
    assert cuda_a["nest"][-12] == 1.0
    assert cuda_a["nest"][-11] == 2.0
    assert cuda_a["nest"][-10] == 3.0
    assert cuda_a["nest"][-9] is None
    assert cuda_a["nest"][-8] is None
    assert cuda_a["nest"][-7] is None
    assert cuda_a["nest"][-6] is None
    assert cuda_a["nest"][-5] == 1.1
    assert cuda_a["nest"][-4] is None
    assert cuda_a["nest"][-3] == 3.3
    assert cuda_a["nest"][-2] is None
    assert cuda_a["nest"][-1] == 5.5
    assert isinstance(cuda_a["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_a.to_typetracer()["nest"][5:].form == cuda_a["nest"][5:].form
    assert len(cuda_a["nest"][5:]) == 8
    assert len(cuda_a["nest"][-8:]) == 8
    assert len(cuda_a["nest"][5:100]) == 8
    assert len(cuda_a["nest"][-8:100]) == 8
    assert cuda_a["nest"][5:][2] is None
    assert cuda_a["nest"][5:][3] == 1.1
    assert cuda_a["nest"][-8:][2] is None
    assert cuda_a["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_a["nest"]["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    cuda_b = ak.to_backend(b, "cuda", highlevel=False)

    assert cuda_b.to_typetracer().form == cuda_b.form
    assert cuda_b.to_typetracer().form.type == cuda_b.form.type
    assert len(cuda_b["nest"]) == 13
    assert cuda_b.to_typetracer()["nest"].form == cuda_b["nest"].form
    with pytest.raises(IndexError):
        cuda_b["nest"][13]
    with pytest.raises(IndexError):
        cuda_b["nest"][-14]
    assert cuda_b["nest"][0] == 0.0
    assert cuda_b["nest"][1] == 1.0
    assert cuda_b["nest"][2] == 2.0
    assert cuda_b["nest"][3] == 3.0
    assert cuda_b["nest"][4] is None
    assert cuda_b["nest"][5] is None
    assert cuda_b["nest"][6] is None
    assert cuda_b["nest"][7] is None
    assert cuda_b["nest"][8] == 1.1
    assert cuda_b["nest"][9] is None
    assert cuda_b["nest"][10] == 3.3
    assert cuda_b["nest"][11] is None
    assert cuda_b["nest"][12] == 5.5
    assert cuda_b["nest"][-13] == 0.0
    assert cuda_b["nest"][-12] == 1.0
    assert cuda_b["nest"][-11] == 2.0
    assert cuda_b["nest"][-10] == 3.0
    assert cuda_b["nest"][-9] is None
    assert cuda_b["nest"][-8] is None
    assert cuda_b["nest"][-7] is None
    assert cuda_b["nest"][-6] is None
    assert cuda_b["nest"][-5] == 1.1
    assert cuda_b["nest"][-4] is None
    assert cuda_b["nest"][-3] == 3.3
    assert cuda_b["nest"][-2] is None
    assert cuda_b["nest"][-1] == 5.5
    assert isinstance(cuda_b["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_b.to_typetracer()["nest"][5:].form == cuda_b["nest"][5:].form
    assert len(cuda_b["nest"][5:]) == 8
    assert len(cuda_b["nest"][-8:]) == 8
    assert len(cuda_b["nest"][5:100]) == 8
    assert len(cuda_b["nest"][-8:100]) == 8
    assert cuda_b["nest"][5:][2] is None
    assert cuda_b["nest"][5:][3] == 1.1
    assert cuda_b["nest"][-8:][2] is None
    assert cuda_b["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_b["nest"]["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    cuda_c = ak.to_backend(c, "cuda", highlevel=False)

    assert cuda_c.to_typetracer().form == cuda_c.form
    assert cuda_c.to_typetracer().form.type == cuda_c.form.type
    assert len(cuda_c["nest"]) == 13
    assert cuda_c.to_typetracer()["nest"].form == cuda_c["nest"].form
    with pytest.raises(IndexError):
        cuda_c["nest"][13]
    with pytest.raises(IndexError):
        cuda_c["nest"][-14]
    assert cuda_c["nest"][0] == 0.0
    assert cuda_c["nest"][1] == 1.0
    assert cuda_c["nest"][2] == 2.0
    assert cuda_c["nest"][3] == 3.0
    assert cuda_c["nest"][4] is None
    assert cuda_c["nest"][5] is None
    assert cuda_c["nest"][6] is None
    assert cuda_c["nest"][7] is None
    assert cuda_c["nest"][8] == 1.1
    assert cuda_c["nest"][9] is None
    assert cuda_c["nest"][10] == 3.3
    assert cuda_c["nest"][11] is None
    assert cuda_c["nest"][12] == 5.5
    assert cuda_c["nest"][-13] == 0.0
    assert cuda_c["nest"][-12] == 1.0
    assert cuda_c["nest"][-11] == 2.0
    assert cuda_c["nest"][-10] == 3.0
    assert cuda_c["nest"][-9] is None
    assert cuda_c["nest"][-8] is None
    assert cuda_c["nest"][-7] is None
    assert cuda_c["nest"][-6] is None
    assert cuda_c["nest"][-5] == 1.1
    assert cuda_c["nest"][-4] is None
    assert cuda_c["nest"][-3] == 3.3
    assert cuda_c["nest"][-2] is None
    assert cuda_c["nest"][-1] == 5.5
    assert isinstance(cuda_c["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_c.to_typetracer()["nest"][5:].form == cuda_c["nest"][5:].form
    assert len(cuda_c["nest"][5:]) == 8
    assert len(cuda_c["nest"][-8:]) == 8
    assert len(cuda_c["nest"][5:100]) == 8
    assert len(cuda_c["nest"][-8:100]) == 8
    assert cuda_c["nest"][5:][2] is None
    assert cuda_c["nest"][5:][3] == 1.1
    assert cuda_c["nest"][-8:][2] is None
    assert cuda_c["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_c["nest"]["bad"]

    # 4.0, 5.0, 6.0, 7.0, 2.2, 4.4, and 6.6 are inaccessible
    d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    cuda_d = ak.to_backend(d, "cuda", highlevel=False)

    assert cuda_d.to_typetracer().form == cuda_d.form
    assert cuda_d.to_typetracer().form.type == cuda_d.form.type
    assert len(cuda_d["nest"]) == 13
    assert cuda_d.to_typetracer()["nest"].form == cuda_d["nest"].form
    with pytest.raises(IndexError):
        cuda_d["nest"][13]
    with pytest.raises(IndexError):
        cuda_d["nest"][-14]
    assert cuda_d["nest"][0] == 0.0
    assert cuda_d["nest"][1] == 1.0
    assert cuda_d["nest"][2] == 2.0
    assert cuda_d["nest"][3] == 3.0
    assert cuda_d["nest"][4] is None
    assert cuda_d["nest"][5] is None
    assert cuda_d["nest"][6] is None
    assert cuda_d["nest"][7] is None
    assert cuda_d["nest"][8] == 1.1
    assert cuda_d["nest"][9] is None
    assert cuda_d["nest"][10] == 3.3
    assert cuda_d["nest"][11] is None
    assert cuda_d["nest"][12] == 5.5
    assert cuda_d["nest"][-13] == 0.0
    assert cuda_d["nest"][-12] == 1.0
    assert cuda_d["nest"][-11] == 2.0
    assert cuda_d["nest"][-10] == 3.0
    assert cuda_d["nest"][-9] is None
    assert cuda_d["nest"][-8] is None
    assert cuda_d["nest"][-7] is None
    assert cuda_d["nest"][-6] is None
    assert cuda_d["nest"][-5] == 1.1
    assert cuda_d["nest"][-4] is None
    assert cuda_d["nest"][-3] == 3.3
    assert cuda_d["nest"][-2] is None
    assert cuda_d["nest"][-1] == 5.5
    assert isinstance(cuda_d["nest"][5:], ak.contents.bytemaskedarray.ByteMaskedArray)
    assert cuda_d.to_typetracer()["nest"][5:].form == cuda_d["nest"][5:].form
    assert len(cuda_d["nest"][5:]) == 8
    assert len(cuda_d["nest"][-8:]) == 8
    assert len(cuda_d["nest"][5:100]) == 8
    assert len(cuda_d["nest"][-8:100]) == 8
    assert cuda_d["nest"][5:][2] is None
    assert cuda_d["nest"][5:][3] == 1.1
    assert cuda_d["nest"][-8:][2] is None
    assert cuda_d["nest"][-8:][3] == 1.1
    with pytest.raises(IndexError):
        cuda_d["nest"]["bad"]


def test_0896_content_classes_refactoring_UnionArray_NumpyArray():
    # 100 is inaccessible in index
    # 1.1 is inaccessible in contents[1]
    a = ak.contents.unionarray.UnionArray.simplified(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3])),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)

    assert cuda_a.to_typetracer().form == cuda_a.form
    assert cuda_a.to_typetracer().form.type == cuda_a.form.type
    assert len(cuda_a) == 7
    with pytest.raises(IndexError):
        cuda_a[7]
    with pytest.raises(IndexError):
        cuda_a[-8]
    assert cuda_a[0] == 5.5
    assert cuda_a[1] == 4.4
    assert cuda_a[2] == 1.0
    assert cuda_a[3] == 2.0
    assert cuda_a[4] == 3.3
    assert cuda_a[5] == 3.0
    assert cuda_a[6] == 5.5
    assert cuda_a[-7] == 5.5
    assert cuda_a[-6] == 4.4
    assert cuda_a[-5] == 1.0
    assert cuda_a[-4] == 2.0
    assert cuda_a[-3] == 3.3
    assert cuda_a[-2] == 3.0
    assert cuda_a[-1] == 5.5
    assert cuda_a.to_typetracer()[3:].form == cuda_a[3:].form
    assert len(cuda_a[3:]) == 4
    assert len(cuda_a[-4:]) == 4
    assert len(cuda_a[3:100]) == 4
    assert len(cuda_a[-4:100]) == 4
    assert cuda_a[3:][1] == 3.3
    assert cuda_a[3:][2] == 3.0
    assert cuda_a[-4:][1] == 3.3
    assert cuda_a[-4:][2] == 3.0
    with pytest.raises(IndexError):
        cuda_a["bad"]
    with pytest.raises(IndexError):
        cuda_a[["bad", "good", "ok"]]


def test_0896_content_classes_refactoring_UnionArray_RecordArray_NumpyArray():
    # 100 is inaccessible in index
    # 1.1 is inaccessible in contents[1]
    a = ak.contents.unionarray.UnionArray.simplified(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak.contents.recordarray.RecordArray(
                [ak.contents.numpyarray.NumpyArray(np.array([1, 2, 3]))], ["nest"]
            ),
            ak.contents.recordarray.RecordArray(
                [
                    ak.contents.numpyarray.NumpyArray(
                        np.array([1.1, 2.2, 3.3, 4.4, 5.5])
                    )
                ],
                ["nest"],
            ),
        ],
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)

    assert cuda_a.to_typetracer().form == cuda_a.form
    assert cuda_a.to_typetracer().form.type == cuda_a.form.type
    assert len(cuda_a["nest"]) == 7
    assert cuda_a.to_typetracer()["nest"].form == cuda_a["nest"].form
    with pytest.raises(IndexError):
        cuda_a["nest"][7]
    with pytest.raises(IndexError):
        cuda_a["nest"][-8]
    assert cuda_a["nest"][0] == 5.5
    assert cuda_a["nest"][1] == 4.4
    assert cuda_a["nest"][2] == 1.0
    assert cuda_a["nest"][3] == 2.0
    assert cuda_a["nest"][4] == 3.3
    assert cuda_a["nest"][5] == 3.0
    assert cuda_a["nest"][6] == 5.5
    assert cuda_a["nest"][-7] == 5.5
    assert cuda_a["nest"][-6] == 4.4
    assert cuda_a["nest"][-5] == 1.0
    assert cuda_a["nest"][-4] == 2.0
    assert cuda_a["nest"][-3] == 3.3
    assert cuda_a["nest"][-2] == 3.0
    assert cuda_a["nest"][-1] == 5.5
    assert cuda_a.to_typetracer()["nest"][3:].form == cuda_a["nest"][3:].form
    assert len(cuda_a["nest"][3:]) == 4
    assert len(cuda_a["nest"][-4:]) == 4
    assert len(cuda_a["nest"][3:100]) == 4
    assert len(cuda_a["nest"][-4:100]) == 4
    assert cuda_a["nest"][3:][1] == 3.3
    assert cuda_a["nest"][3:][2] == 3.0
    assert cuda_a["nest"][-4:][1] == 3.3
    assert cuda_a["nest"][-4:][2] == 3.0
    with pytest.raises(IndexError):
        cuda_a["nest"]["bad"]


def test_2426_is_equal_to_equal_union():
    union_1 = ak.from_iter([1, None, {"x": 2}], highlevel=False)
    union_2 = ak.from_iter([1, None, {"x": 2}], highlevel=False)

    cuda_union_1 = ak.to_backend(union_1, "cuda", highlevel=False)
    cuda_union_2 = ak.to_backend(union_2, "cuda", highlevel=False)

    assert cuda_union_1.is_equal_to(cuda_union_2)


def test_2426_is_equal_to_unequal_union():
    union_1 = ak.from_iter([1, None, {"x": 2}, 3], highlevel=False)
    union_2 = ak.from_iter([1, None, {"x": 2}, 2], highlevel=False)

    cuda_union_1 = ak.to_backend(union_1, "cuda", highlevel=False)
    cuda_union_2 = ak.to_backend(union_2, "cuda", highlevel=False)

    assert not cuda_union_1.is_equal_to(cuda_union_2)


def test_0198_tutorial_documentation_1_firsts():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(
        ak.operations.firsts(
            ak.operations.singletons(cuda_array),
            axis=1,
        )
    ) == [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5]

    array = ak.Array([[1.1, 2.2, None], [3.3, None], [None], [4.4, 5.5]])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(
        ak.operations.firsts(
            ak.operations.singletons(cuda_array, axis=1),
            axis=2,
        )
    ) == [[1.1, 2.2, None], [3.3, None], [None], [4.4, 5.5]]

    array = ak.Array([[[1.1, 2.2, None]], [[3.3, None]], [[None]], [[4.4, 5.5]]])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(
        ak.operations.firsts(
            ak.operations.singletons(
                cuda_array,
                axis=2,
            ),
            axis=3,
        )
    ) == [[[1.1, 2.2, None]], [[3.3, None]], [[None]], [[4.4, 5.5]]]


def test_0198_tutorial_documentation_1_singletons():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(ak.operations.singletons(cuda_array)) == [
        [1.1],
        [2.2],
        [],
        [3.3],
        [],
        [],
        [4.4],
        [5.5],
    ]
    assert to_list(ak.operations.singletons(cuda_array)) == [
        [1.1],
        [2.2],
        [],
        [3.3],
        [],
        [],
        [4.4],
        [5.5],
    ]

    array = ak.Array([[1.1, 2.2, None], [3.3, None], [None], [4.4, 5.5]])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(ak.operations.singletons(cuda_array, axis=1)) == [
        [[1.1], [2.2], []],
        [[3.3], []],
        [[]],
        [[4.4], [5.5]],
    ]

    array = ak.Array([[[1.1, 2.2, None]], [[3.3, None]], [[None]], [[4.4, 5.5]]])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(
        ak.operations.singletons(
            cuda_array,
            axis=2,
        )
    ) == [[[[1.1], [2.2], []]], [[[3.3], []]], [[[]]], [[[4.4], [5.5]]]]


def test_0198_tutorial_documentation_1_allow_missing():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    cuda_array = ak.to_backend(array, "cuda")

    ak.operations.to_numpy(cuda_array)
    with pytest.raises(ValueError):
        ak.operations.to_numpy(cuda_array, allow_missing=False)


def test_0198_tutorial_documentation_1_flatten0():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(ak.operations.flatten(cuda_array, axis=0)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]

    content0 = ak.operations.from_iter(
        [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5], highlevel=False
    )
    content1 = ak.operations.from_iter(
        ["one", None, "two", None, "three"], highlevel=False
    )
    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8(
                np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0], dtype=np.int8)
            ),
            ak.index.Index64(
                np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 5, 6, 4, 7], dtype=np.int64)
            ),
            [content0, content1],
        )
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == [
        1.1,
        "one",
        2.2,
        None,
        None,
        "two",
        3.3,
        None,
        None,
        None,
        4.4,
        "three",
        5.5,
    ]
    assert to_list(ak.operations.flatten(cuda_array, axis=0)) == [
        1.1,
        "one",
        2.2,
        "two",
        3.3,
        4.4,
        "three",
        5.5,
    ]


def test_0334_fully_broadcastable_where():
    one = ak.highlevel.Array([[0, 1, 2], [], [3, 4], [5], [6, 7, 8, 9]])
    two = ak.highlevel.Array(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    )
    condition = ak.highlevel.Array(
        [[False, True, False], [], [True, False], [True], [False, False, True, True]]
    )

    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")
    cuda_condition = ak.to_backend(condition, "cuda")

    assert ak.operations.where(cuda_condition, cuda_one, cuda_two).to_list() == [
        [0, 1, 2.2],
        [],
        [3, 4.4],
        [5],
        [6.6, 7.7, 8, 9],
    ]


def test_0334_fully_broadcastable_where_issue_334():
    a = ak.highlevel.Array([1, 2, 3, 4])
    b = ak.highlevel.Array([-1])
    c = ak.highlevel.Array([True, False, True, True])

    cuda_a = ak.to_backend(a, "cuda")
    cuda_b = ak.to_backend(b, "cuda")
    cuda_c = ak.to_backend(c, "cuda")

    assert ak.operations.where(cuda_c, cuda_a, cuda_b).to_list() == [1, -1, 3, 4]
    assert ak.operations.where(
        *ak.operations.broadcast_arrays(cuda_c, cuda_a, cuda_b)
    ).to_list() == [
        1,
        -1,
        3,
        4,
    ]
    assert ak.operations.where(cuda_c, cuda_a, -1).to_list() == [1, -1, 3, 4]
    assert ak.operations.where(
        *ak.operations.broadcast_arrays(cuda_c, cuda_a, -1)
    ).to_list() == [
        1,
        -1,
        3,
        4,
    ]


def test_0150_flatten_axis_none_0866():
    a1 = ak.operations.zip(
        {"a": [[1], [], [2, 3]], "b": [[4], [], [5, 6]]}, with_name="a1"
    )
    a2 = ak.operations.zip(
        {"a": [[7, 8], [9], []], "b": [[10, 11], [12], []]}, with_name="a2"
    )
    cuda_a1 = ak.to_backend(a1, "cuda")
    cuda_a2 = ak.to_backend(a2, "cuda")

    condition = ak.Array([True, False, True])
    cuda_condition = ak.to_backend(condition, "cuda")
    union = ak.operations.where(cuda_condition, cuda_a1, cuda_a2)

    assert set(cp.asnumpy(ak.operations.flatten(union, axis=None))) == {
        1,
        2,
        3,
        4,
        5,
        6,
        9,
        12,
    }


def test_0150_flatten_0198():
    array = ak.highlevel.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5]).layout
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=0)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]

    content0 = ak.operations.from_iter(
        [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5], highlevel=False
    )
    content1 = ak.operations.from_iter(
        ["one", None, "two", None, "three"], highlevel=False
    )
    array = ak.contents.UnionArray(
        ak.index.Index8(
            np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0], dtype=np.int8)
        ),
        ak.index.Index64(
            np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 5, 6, 4, 7], dtype=np.int64)
        ),
        [content0, content1],
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == [
        1.1,
        "one",
        2.2,
        None,
        None,
        "two",
        3.3,
        None,
        None,
        None,
        4.4,
        "three",
        5.5,
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=0)) == [
        1.1,
        "one",
        2.2,
        "two",
        3.3,
        4.4,
        "three",
        5.5,
    ]


def test_0150_flatten_0042():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=1)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]


def test_2678_same_backend_where():
    result = ak.where(
        ak.to_backend([True, False, False], "cuda"),
        ak.to_backend([1, 2, 3], "cuda"),
        ak.to_backend([4, 5, 6], "cuda"),
    )

    assert ak.backend(result) == "cuda"


behavior_1 = {"foo": "bar"}
behavior_2 = {"baz": "bargh!"}
behavior = {**behavior_1, **behavior_2}


@pytest.mark.parametrize(
    ("func", "axis"),
    [
        (ak.std, -1),
        (ak.var, -1),
        (ak.softmax, -1),
        *product(
            (
                ak.any,
                ak.min,
                ak.sum,
                ak.ptp,
                ak.count_nonzero,
                lambda *args, **kwargs: ak.moment(*args, **kwargs, n=3),
                ak.all,
                ak.mean,
                ak.max,
                ak.prod,
                ak.count,
            ),
            ([-1]),
        ),
    ],
)
def test_2754_highlevel_behavior_missing_reducers(axis, func):
    array = ak.Array([[1, 2, 3, 4], [5], [10]])
    cuda_array = ak.to_backend(array, "cuda", behavior=behavior_1)

    assert (
        func(
            cuda_array,
            axis=axis,
            highlevel=True,
            behavior=behavior_2,
        ).behavior
        == behavior_2
    )
    assert (
        func(
            cuda_array,
            axis=axis,
            highlevel=True,
        ).behavior
        == behavior_1
    )
    del cuda_array


def test_BitMaskedArray():
    content = ak.contents.NumpyArray(np.arange(13))
    mask = ak.index.IndexU8(np.array([58, 59], dtype=np.uint8))
    array = ak.contents.BitMaskedArray(
        mask, content, valid_when=False, length=13, lsb_order=False
    )
    array = ak.to_backend(array, "cuda", highlevel=False)
    assert cp.asarray(array.mask_as_bool(valid_when=True)).tolist() == [
        True,
        True,
        False,
        False,
        False,
        True,
        False,
        True,
        True,
        True,
        False,
        False,
        False,
    ]
    assert cp.asarray(array.to_ByteMaskedArray().mask).tolist() == [
        0,
        0,
        1,
        1,
        1,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
    ]
    assert cp.asarray(array.to_IndexedOptionArray64().index).tolist() == [
        0,
        1,
        -1,
        -1,
        -1,
        5,
        -1,
        7,
        8,
        9,
        -1,
        -1,
        -1,
    ]
    assert to_list(array) == [
        0,
        1,
        None,
        None,
        None,
        5,
        None,
        7,
        8,
        9,
        None,
        None,
        None,
    ]
    del array



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak


def test_to_numpy_record_array():
    test_array = ak.Array([{"x": 1, "y": 1.1}, None]).to_numpy()
    expected_numpy = np.ma.masked_array(
        data=[(1, 1.1), (np.iinfo(np.int64).max, np.nan)],
        mask=[(False, False), (True, True)],
        dtype=[("x", "<i8"), ("y", "<f8")],
    )
    assert np.array_equal(test_array["x"], expected_numpy["x"])
    # equal_nan not supported on integer arrays
    try:
        assert np.array_equal(test_array["y"], expected_numpy["y"], equal_nan=True)
    except TypeError:
        # older numpy versions do not support `equal_nan`
        assert np.array_equal(test_array["y"].filled(), expected_numpy["y"].filled())



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    array = ak.Array([[True, False, True], [False, False, True, False], [True, False]])

    assert ak.argmin(array, axis=1).tolist() == [1, 0, 1]
    assert ak.argmax(array, axis=1).tolist() == [0, 2, 0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pathlib

import pytest

import awkward

pytest.importorskip("pyarrow.parquet")

metadata_from_parquet_submodule = pytest.importorskip(
    "awkward.operations.ak_metadata_from_parquet"
)
metadata_from_parquet = metadata_from_parquet_submodule.metadata_from_parquet

SAMPLES_DIR = pathlib.Path(__file__).parent / "samples"
input = SAMPLES_DIR / "nullable-record-primitives.parquet"


def test_parquet_uuid():
    meta = metadata_from_parquet(input)
    assert (
        meta["uuid"]
        == "582dabdb8c87bfa17bc930676ed26b8d4ab22a900f92357751dc380c41acb593"
    )


@pytest.mark.parametrize("calculate_uuid", [True, False])
def test_return_tuple_with_or_without_uuid(calculate_uuid):
    results = awkward.operations.ak_from_parquet.metadata(
        input,
        {},
        None,
        None,
        False,
        True,
        calculate_uuid=calculate_uuid,
    )
    if calculate_uuid:
        assert len(results) == 8, "Expected 8 items in the result tuple"
        (
            _parquet_columns,
            _subform,
            _actual_paths,
            _fs,
            _subrg,
            _col_counts,
            _metadata,
            uuid,
        ) = results
        assert uuid is not None, "UUID should be present when calculate_uuid is True"
        print("uuid:", uuid)
    else:
        assert len(results) == 7, "Expected 7 items in the result tuple"
        (
            _parquet_columns,
            _subform,
            _actual_paths,
            _fs,
            _subrg,
            _col_counts,
            _metadata,
        ) = results



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak


def test_wrap_index_cupy():
    cp = pytest.importorskip("cupy")
    data = cp.arange(10, dtype=cp.int64)
    index = ak.index.Index64(data)
    other_data = cp.asarray(index)
    result = cp.shares_memory(data, other_data)
    assert result is True



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak
from awkward._nplikes.array_like import MaterializableArray
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.placeholder import PlaceholderArray
from awkward._nplikes.virtual import VirtualNDArray


def test():
    numpy = Numpy.instance()
    a = ak.contents.NumpyArray(np.arange(10, dtype=np.float32))
    b = ak.contents.NumpyArray(PlaceholderArray(numpy, (10,), np.float32))
    layout = ak.contents.RecordArray([a, b], None, 5)
    layout.materialize(VirtualNDArray)
    layout.to_packed()
    with pytest.raises(RuntimeError, match="should never have been encountered"):
        layout.materialize(MaterializableArray)

    a = ak.contents.NumpyArray(
        VirtualNDArray(
            numpy, (10,), np.float32, lambda: np.arange(10, dtype=np.float32)
        )
    )
    b = ak.contents.NumpyArray(PlaceholderArray(Numpy.instance(), (10,), np.float32))
    layout = ak.contents.RecordArray([a, b], None, 5)
    layout.materialize(VirtualNDArray)
    layout.to_packed()
    with pytest.raises(RuntimeError, match="should never have been encountered"):
        layout.materialize(MaterializableArray)



# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_1():
    tocarry = cupy.array([123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    slicestops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [2, 0, 1, 2, 3]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_2():
    tocarry = cupy.array([123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    slicestops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [2, 0, 1, 2, 3]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_3():
    tocarry = cupy.array([123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    slicestops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [2, 0, 1, 2, 3]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_4():
    tocarry = cupy.array([123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    slicestops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [2, 0, 1, 2, 3]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_5():
    tocarry = cupy.array([123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    slicestops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    length = 3
    missing = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [2, 0, 1, 2, 3]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [2, 3, 5, 7]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_6():
    tocarry = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    slicestops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 0, 1, 2, 3, 4]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_7():
    tocarry = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    slicestops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 0, 1, 2, 3, 4]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_8():
    tocarry = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    slicestops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 0, 1, 2, 3, 4]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_9():
    tocarry = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    slicestops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 0, 1, 2, 3, 4]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_10():
    tocarry = cupy.array([123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    slicestops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    length = 3
    missing = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 0, 1, 2, 3, 4]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 8, 12, 17]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_11():
    tocarry = cupy.array([123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    slicestops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [7, 8]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_12():
    tocarry = cupy.array([123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    slicestops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [7, 8]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_13():
    tocarry = cupy.array([123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    slicestops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [7, 8]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_14():
    tocarry = cupy.array([123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    slicestops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [7, 8]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_15():
    tocarry = cupy.array([123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    slicestops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    length = 3
    missing = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [7, 8]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [1, 1, 3, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_16():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    slicestops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 0, 0]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_17():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    slicestops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 0, 0]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_18():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    slicestops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 0, 0]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_19():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    slicestops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    missing = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 0, 0]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))

def test_cudaawkward_ListArray_getitem_jagged_shrink_64_20():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    tosmalloffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    tolargeoffsets = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    slicestarts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    slicestops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    length = 3
    missing = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    funcC = cupy_backend['awkward_ListArray_getitem_jagged_shrink', cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(tocarry, tosmalloffsets, tolargeoffsets, slicestarts, slicestops, length, missing)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 0, 0]
    cpt.assert_allclose(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))
    pytest_tosmalloffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tosmalloffsets[:len(pytest_tosmalloffsets)], cupy.array(pytest_tosmalloffsets))
    pytest_tolargeoffsets = [0, 1, 2, 3]
    cpt.assert_allclose(tolargeoffsets[:len(pytest_tolargeoffsets)], cupy.array(pytest_tolargeoffsets))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np

import awkward as ak
from awkward.types import ListType
from awkward.types.numpytype import NumpyType


def test_complex_ops():
    """
    Testing Awkward implementations (cpu backend) of real, imag, and angle
    """
    arr_1 = ak.Array([[1 + 0.1j, 2 + 0.2j, 3 + 0.3j], [], [4 + 0.4j, 5 + 0.5j]])
    arr_csingle = ak.from_numpy(np.array([6.0 + 0.0j, 0.0 + 7.0j], dtype="complex64"))
    arr_real = ak.Array([11.0, 12.0])

    real_1 = np.real(arr_1)
    assert ak.all(real_1 == ak.Array([[1, 2, 3], [], [4, 5]]))

    real_csingle = np.real(arr_csingle)
    assert ak.all(real_csingle == ak.Array([6.0, 0.0]))
    assert real_csingle.type.content == NumpyType("float32")

    real_real = np.real(arr_real)
    assert ak.all(real_real == ak.Array([11.0, 12.0]))

    imag_1 = np.imag(arr_1)
    assert ak.all(imag_1 == ak.Array([[0.1, 0.2, 0.3], [], [0.4, 0.5]]))

    imag_csingle = np.imag(arr_csingle)
    assert ak.all(imag_csingle == ak.Array([0.0, 7.0]))
    assert imag_csingle.type.content == NumpyType("float32")

    imag_real = np.imag(arr_real)
    assert ak.all(imag_real == ak.Array([0.0, 0.0]))

    angle_1 = np.angle(arr_1)
    a1 = np.arctan(0.1)
    expected = ak.Array([[a1, a1, a1], [], [a1, a1]])  # or [[a1] * 3, ...]
    assert ak.all(np.abs(angle_1 - expected) < np.finfo(np.float64).eps)

    angle_csingle = np.angle(arr_csingle, deg=True)
    assert ak.all(angle_csingle == ak.Array([0.0, 90.0]))
    assert angle_csingle.type.content == NumpyType("float32")

    angle_real = np.angle(arr_real)
    assert ak.all(angle_real == ak.Array([0.0, 0.0]))


def test_complex_typetracer():
    tt_arr = ak.to_backend(
        ak.Array([[1 + 0.1j, 2 + 0.2j, 3 + 0.3j], [], [4 + 0.4j, 5 + 0.5j]]),
        "typetracer",
    )
    tt_csingle = ak.to_backend(
        ak.from_numpy(np.array([6.0 + 0.0j, 0.0 + 7.0j], dtype="complex64")),
        "typetracer",
    )
    tt_real = ak.to_backend(
        ak.from_numpy(np.array([11.0, 12.0], dtype="float16")), "typetracer"
    )
    tt_int = ak.to_backend(ak.Array([1, 2]), "typetracer")

    real_arr = np.real(tt_arr)
    assert real_arr.type.content == ListType(NumpyType("float64"))
    real_tt_csingle = np.real(tt_csingle)
    assert real_tt_csingle.type.content == NumpyType("float32")
    real_tt_real = np.real(tt_real)
    assert real_tt_real.type.content == NumpyType("float16")

    imag_tt_csingle = np.imag(tt_csingle)
    assert imag_tt_csingle.type.content == NumpyType("float32")
    imag_tt_real = np.imag(tt_real)
    assert imag_tt_real.type.content == NumpyType("float16")

    angle_tt_csingle = np.angle(tt_csingle, deg=True)
    assert angle_tt_csingle.type.content == NumpyType("float32")
    angle_tt_real = np.angle(tt_real)
    assert angle_tt_real.type.content == NumpyType("float16")
    assert np.angle(tt_int).type.content == NumpyType("float64")



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak
from awkward.types import ArrayType, NumpyType, RegularType

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0150_ByteMaskedArray_flatten():
    content = ak.operations.from_iter(
        [
            [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 7.7, 8.8, 9.9]],
            [[], [10.0, 11.1, 12.2]],
        ],
        highlevel=False,
    )
    mask = ak.index.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.contents.ByteMaskedArray(mask, content, valid_when=False)
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=1)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=-2)) == [
        [0.0, 1.1, 2.2],
        [],
        [3.3, 4.4],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=2)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        None,
        None,
        [10.0, 11.1, 12.2],
    ]
    assert ak.operations.to_list(ak.operations.flatten(cuda_array, axis=-1)) == [
        [0.0, 1.1, 2.2, 3.3, 4.4],
        [],
        None,
        None,
        [10.0, 11.1, 12.2],
    ]


def test_1586_should_preserve_regulararray_numpy_regular_axis1():
    a1 = ak.Array(np.array([[0.0, 1.1], [2.2, 3.3]]))
    a2 = ak.from_json("[[4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]")
    cuda_a1 = ak.to_backend(a1, "cuda")
    assert isinstance(cuda_a1.layout, ak.contents.NumpyArray)

    a2 = ak.to_regular(a2, axis=1)
    cuda_a2 = ak.to_backend(a2, "cuda")

    c = ak.concatenate([cuda_a1, cuda_a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 5), 2)


def test_1586_should_preserve_regulararray_regular_numpy_axis1():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.Array(np.array([[4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]))

    cuda_a1 = ak.to_backend(a1, "cuda")
    cuda_a2 = ak.to_backend(a2, "cuda")

    cuda_a1 = ak.to_regular(cuda_a1, axis=1)

    assert isinstance(cuda_a2.layout, ak.contents.NumpyArray)
    c = ak.concatenate([cuda_a1, cuda_a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 5), 2)


def test_1586_should_preserve_regulararray_regular_regular_axis1():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]")

    cuda_a1 = ak.to_backend(a1, "cuda")
    cuda_a2 = ak.to_backend(a2, "cuda")

    cuda_a1 = ak.to_regular(cuda_a1, axis=1)
    cuda_a2 = ak.to_regular(cuda_a2, axis=1)
    c = ak.concatenate([cuda_a1, cuda_a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 5), 2)


def test_0072_regulararray_fillna_unionarray():
    content1 = ak.operations.from_iter([[], [1.1], [2.2, 2.2]], highlevel=False)
    content2 = ak.operations.from_iter([["two", "two"], ["one"], []], highlevel=False)
    tags = ak.index.Index8(np.array([0, 1, 0, 1, 0, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2], dtype=np.int64))
    array = ak.contents.UnionArray(tags, index, [content1, content2])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array) == [[], ["two", "two"], [1.1], ["one"], [2.2, 2.2], []]

    padded_array = ak._do.pad_none(array, 2, 1)
    assert to_list(padded_array) == [
        [None, None],
        ["two", "two"],
        [1.1, None],
        ["one", None],
        [2.2, 2.2],
        [None, None],
    ]

    value = ak.contents.NumpyArray(np.array([777]))

    assert to_list(ak._do.fill_none(padded_array, value)) == [
        [777, 777],
        ["two", "two"],
        [1.1, 777],
        ["one", 777],
        [2.2, 2.2],
        [777, 777],
    ]


def test_0590_allow_regulararray_size_zero_ListOffsetArray_rpad_and_clip():
    array = ak.highlevel.Array([[1, 2, 3], [], [4, 5]])
    assert ak.operations.pad_none(array, 0, clip=True).to_list() == [
        [],
        [],
        [],
    ]

    array = ak.highlevel.Array([[1, 2, 3], [], [4, 5]])
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.pad_none(cuda_array, 0).to_list() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def check(layout, assert_length):
    if hasattr(layout, "contents"):
        for x in layout.contents:
            check(x, assert_length)
    elif hasattr(layout, "content"):
        check(layout.content, assert_length)
    else:
        assert layout.length <= assert_length


def test_2arrays():
    one_a = ak.Array([{"x": 1, "y": 2}], with_name="T")
    one_b = ak.Array([{"x": 1, "y": 2}], with_name="T")
    two_a = ak.Array([{"x": 1, "z": 3}], with_name="T")
    two_b = ak.Array([{"x": 1, "z": 3}], with_name="T")
    three = ak.Array([{"x": 4}, {"x": 4}], with_name="T")

    first = ak.zip({"a": one_a, "b": one_b})
    second = ak.zip({"a": two_a, "b": two_b})

    cat = ak.concatenate([first, second], axis=0)

    cat["another"] = three

    for _ in range(5):
        check(cat.layout, 2)

        cat["another", "w"] = three.x


def test_3arrays():
    zero_a = ak.Array([{"x": 1, "y": 1}], with_name="T")
    zero_b = ak.Array([{"x": 1, "v": 1}], with_name="T")
    one_a = ak.Array([{"x": 1, "y": 2}], with_name="T")
    one_b = ak.Array([{"x": 1, "y": 2}], with_name="T")
    two_a = ak.Array([{"x": 1, "z": 3}], with_name="T")
    two_b = ak.Array([{"x": 1, "z": 3}], with_name="T")
    three = ak.Array([{"x": 4}, {"x": 4}, {"x": 4}], with_name="T")

    zeroth = ak.zip({"a": zero_a, "b": zero_b})
    first = ak.zip({"a": one_a, "b": one_b})
    second = ak.zip({"a": two_a, "b": two_b})

    cat = ak.concatenate([zeroth, first, second], axis=0)

    cat["another"] = three

    for _ in range(5):
        check(cat.layout, 3)

        cat["another", "w"] = three.x



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    arr = ak.Array([[1], [2, 3], [1, 2, 4, 5]])[[0, None, 2]]
    l1 = ak.typetracer.length_one_if_typetracer(ak.to_backend(arr, "typetracer"))

    assert l1.to_list() == [None]
    assert str(l1.type) == "1 * option[var * int64]"
